// forint.cc
//
//

#include <forint/forint.hh>

namespace FInt {

namespace FInt_internal {

static double const
constant [] = { M_E, M_LOG2E, M_LOG10E, M_LN2, M_LN10,
		M_PI, M_PI_2, M_PI_4, M_1_PI, M_2_PI, M_2_SQRTPI,
		M_SQRT2, M_SQRT1_2 };

// unary functions("FUNC")...
double abs   (double a) { return fabs(int(a)); }
double labs  (double a) { return fabs(long(a)); }
double not   (double a) { return !a; }
double comp  (double a) { return ~(Llong(a)); }
double u_min (double a) { return -a; }
typedef double (*FUNC_PTR) (double);
FUNC_PTR func_arr [] = { sin, cos, tan, asin, acos, atan,
			 sinh, cosh, tanh, asinh, acosh, atanh,
			 exp, log, log10, sqrt,
			 expm1, log1p, cbrt, j0, j1, y0, y1, erf, erfc, lgamma,
			 fabs, abs, labs, ceil, floor, rint, not, comp, u_min};

// binary functions("OPER")...
double add  (double a, double b) { return a + b; }
double subt (double a, double b) { return a - b; }
double mult (double a, double b) { return a * b; }
double div  (double a, double b) { return a / b; }
double mod  (double a, double b) { return Llong(a) % Llong(b); }
double band (double a, double b) { return Llong(a) & Llong(b); }
double bor  (double a, double b) { return Llong(a) | Llong(b); }
double bxor (double a, double b) { return Llong(a) ^ Llong(b); }
double loff (double a, double b) { return Llong(a) << Llong(b); }
double roff (double a, double b) { return Llong(a) >> Llong(b); }
double ldexp (double a, double b) { return ldexp(a, int(b)); }
double jn (double a, double b) { return jn(int(a),b); }
double yn (double a, double b) { return yn(int(a),b); }
typedef double (*OPER_PTR) (double,double);
OPER_PTR oper_arr [] = { add, subt, mult, div, mod,
			 band, bor, bxor, loff, roff,
			 atan2, fmod, pow, hypot,
			 ldexp, drem, copysign, jn, yn };

// conditions("COND")...
bool gt (double a, double b) { return a >  b; }
bool ge (double a, double b) { return a >= b; }
bool lt (double a, double b) { return a <  b; }
bool le (double a, double b) { return a <= b; }
bool eq (double a, double b) { return a == b; }
bool ne (double a, double b) { return a != b; }
typedef bool (*COND_PTR) (double,double);
COND_PTR cond_arr [] = { gt, ge, lt, le, eq, ne };

// logic("LOGC")...
/*
bool and (bool a, bool b) { return a && b; }
bool or  (bool a, bool b) { return a || b; }
typedef bool (*LOGC_PTR) (bool,bool);
LOGC_PTR logc_arr [] = { and, or };
*/

// assignment...
double ass     (double* a, double b) { return *a  = b; }
double ass_pl  (double* a, double b) { return *a += b; }
double ass_min (double* a, double b) { return *a -= b; }
double ass_mul (double* a, double b) { return *a *= b; }
double ass_div (double* a, double b) { return *a /= b; }
double ass_mod (double* a, double b) { return *a = Llong(*a) % Llong(b); }
double ass_and (double* a, double b) { return *a = Llong(*a) & Llong(b); }
double ass_or  (double* a, double b) { return *a = Llong(*a) | Llong(b); }
double ass_xor (double* a, double b) { return *a = Llong(*a) ^ Llong(b); }
double ass_loff(double* a, double b) { return *a = Llong(*a) << Llong(b); }
double ass_roff(double* a, double b) { return *a = Llong(*a) >> Llong(b); }
typedef double (*ASS_PTR) (double*,double);
ASS_PTR ass_arr [] = { ass, ass_pl, ass_min, ass_mul, ass_div, ass_mod,
		       ass_and, ass_or, ass_xor, ass_loff, ass_roff };

}

// make calculation by formula in a mode:
//   RVM_MODE - by a recurcive calls virtual machine,
//   SVM_MODE - by a stack virtual machine,
//   DLF_MODE - by a dynamic linking of a compiled formula, in the mode
//              external names must coinside with their aliases !!! ...
double forint::
calculate(calc_mode_t mode)
{
  error_code = 0;
  break_flag = false;
  if (!r_action.size() || !s_action.size()) {
if (!locale) {
    c.err(0,0,"try to run empty program.\n");
} else {
    c.err(0,0,"    .\n");
}
    c.flsh(); 
    error_code = 1010;
    return error_code;
  }
  using namespace FInt_internal;
  number_p = number.getPtr();
  offsets_p = offsets.getPtr();
#ifdef FINT_DEBUG_PRINT
  cout << "\noffsets: ";
  offsets.show(); cout << endl;
#endif
  var_num_p     = var_num.getPtr();
  poin_num_p    = poin_num.getPtr();
  poin_num_level_p    = poin_num_level.getPtr();
  usr_function_0_num_p = usr_function_0_num.getPtr();
  usr_function_1_num_p = usr_function_1_num.getPtr();
  usr_function_2d_num_p= usr_function_2d_num.getPtr();
  usr_function_2p_num_p= usr_function_2p_num.getPtr();
  usr_function_2_num_p = usr_function_2_num.getPtr();
  register int k = 0;
  register double  value = 0;
  register void* p_value = 0;

  switch (mode) {
  case RVM_MODE:
    r_action_p = r_action.getPtr();
    do {
      if (r_calculate(k, value, p_value)) {
if (!locale) {
	c.err(0,0,"forint internal error.\n");
} else {
	c.err(0,0,"  forint.\n");
}
        c.flsh(); 
	error_code = -100;
	return error_code;
      }
    } while (++k < int(r_action.size()));
    break;
  case SVM_MODE:
    s_action_p = s_action.getPtr();
    if (s_action.size() && s_calculate(k, value, s_action.size())) {
if (!locale) {
      c.err(0,0,"forint internal error.\n");
} else {
      c.err(0,0,"  forint.\n");
}
      c.flsh(); 
      error_code = -101;
      return error_code;
    }
    break;
  case DLF_MODE:
    if (dl_calculate(value, math_expression))
      return error_code;
    break;
  default:;
  }
  return value;
}


// a recursive call virtual machine...
int forint::
r_calculate(int&   i,        // start with i-th command in the "r_action"
	    double& value,   // returned value
	    void*&  p_value)
{
  using namespace FInt_internal;
  register double val1 = 0;
  register double val2 = 0;
  register void* pval  = 0;
  register void* pval2 = 0;
  register long j = 0;
  register char*   t1 = 0;
  register short*  t2 = 0;
  register int*    t3 = 0;
  register long*   t4 = 0;
  register long long* t41= 0;
  register float*  t5 = 0;
  register double* t6 = 0;
  register long double* t7 = 0;
  register bool cont_flag = false;
  register bool non_lvalue = true;
  register uint n = 0;
  register bool cond;

 cont:
  if (r_action_p[i] == FI_NUMB) {
    non_lvalue = false;
    value = *number_p++;
  }
  else if (r_action_p[i] == FI_VRBL) {
    non_lvalue = false;
    p_value = variable_p[*var_num_p];
    value = *variable_p[*var_num_p++];
  }
  else if (r_action_p[i] == FI_POIN) {
    non_lvalue = false;
    uint* pp = poin_offset.getPtr()[*poin_num_p];
    uint* ppp = poin_dim_size.getPtr()[*poin_num_p];

    if (poin_type.getPtr()[*poin_num_p].level) {
      void** t = (void**)pointer.getPtr()[*poin_num_p];
      j = *poin_num_level_p++;
      for (; j-- > 0;) {
	r_calculate(++i, val1, pval);
	if (val1 > *ppp || val1 < 0) {
	  out_of_bound(poin_name.getPtr()[*poin_num_p],*ppp,int(val1));
	  break;
	}
	ppp++;
	t = t + (int)val1 * *pp++;
      }
      p_value = t;
    } else
      switch (poin_type.getPtr()[*poin_num_p].b_type) {

      case FI_CHAR:
	t1 = (char*)pointer.getPtr()[*poin_num_p];
	j = *poin_num_level_p++;
	for (; j-- > 0;) {
	  r_calculate(++i, val1, pval);
	  if (val1 > *ppp || val1 < 0) {
	    out_of_bound(poin_name.getPtr()[*poin_num_p],*ppp,int(val1));
	    break;
	  }
	  ppp++;
	  t1 += (int)val1 * *pp++;
	}
	p_value = t1;
	value = *t1;
	break;

      case FI_SHORT:
	t2 = (short*)pointer.getPtr()[*poin_num_p];
	j = *poin_num_level_p++;
	for (; j-- > 0;) {
	  r_calculate(++i,val1, pval);
	  if (val1 > *ppp || val1 < 0) {
	    out_of_bound(poin_name.getPtr()[*poin_num_p],*ppp,int(val1));
	    break;
	  }
	  ppp++;
	  t2 += (int)val1 * *pp++;
	}
	p_value = t2;
	value = *t2;
	break;

      case FI_INT:
	t3 = (int*)pointer.getPtr()[*poin_num_p];
	j = *poin_num_level_p++;
	for (; j-- > 0;) {
	  r_calculate(++i, val1, pval);
	  if (val1 > *ppp || val1 < 0) {
	    out_of_bound(poin_name.getPtr()[*poin_num_p],*ppp,int(val1));
	    break;
	  }
	  ppp++;
	  t3 += (int)val1 * *pp++;
	}
	p_value = t3;
	value = *t3;
	break;

      case FI_LONG:
	t4 = (long*)pointer.getPtr()[*poin_num_p];
	j = *poin_num_level_p++;
	for (; j-- > 0;) {
	  r_calculate(++i, val1, pval);
	  if (val1 > *ppp || val1 < 0) {
	    out_of_bound(poin_name.getPtr()[*poin_num_p],*ppp,int(val1));
	    break;
	  }
	  ppp++;
	  t4 += (int)val1 * *pp++;
	}
	p_value = t4;
	value = *t4;
	break;

      case FI_LONG_LONG:
	t41 = (long long*)pointer.getPtr()[*poin_num_p];
	j = *poin_num_level_p++;
	for (; j-- > 0;) {
	  r_calculate(++i, val1, pval);
	  if (val1 > *ppp || val1 < 0) {
	    out_of_bound(poin_name.getPtr()[*poin_num_p],*ppp,int(val1));
	    break;
	  }
	  ppp++;
	  t41 += (int)val1 * *pp++;
	}
	p_value = t41;
	value = *t41;
	break;

      case FI_FLOAT:
	t5 = (float*)pointer.getPtr()[*poin_num_p];
	j = *poin_num_level_p++;
	for (; j-- > 0;) {
	  r_calculate(++i, val1, pval);
	  if (val1 > *ppp || val1 < 0) {
	    out_of_bound(poin_name.getPtr()[*poin_num_p],*ppp,int(val1));
	    break;
	  }
	  ppp++;
	  t5 += (int)val1 * *pp++;
	}
	p_value = t5;
	value = *t5;
	break;

      case FI_DOUBLE:
	t6 = (double*)pointer.getPtr()[*poin_num_p];
	j = *poin_num_level_p++;
	for (; j-- > 0;) {
	  r_calculate(++i, val1, pval);
	  if (val1 > *ppp || val1 < 0) {
	    out_of_bound(poin_name.getPtr()[*poin_num_p],*ppp,int(val1));
	    break;
	  }
	  ppp++;
	  t6 += (int)val1 * *pp++;
	}
	p_value = t6;
	value = *t6;
	break;

      case FI_LONG_DOUBLE:
	t7 = (long double*)pointer.getPtr()[*poin_num_p];
	j = *poin_num_level_p++;
	for (; j-- > 0;) {
	  r_calculate(++i, val1, pval);
	  if (val1 > *ppp || val1 < 0) {
	    out_of_bound(poin_name.getPtr()[*poin_num_p],*ppp,int(val1));
	    break;
	  }
	  ppp++;
	  t7 += (int)val1 * *pp++;
	}
	p_value = t7;
	value = *t7;
	break;

	case FI_USR_CLASS:
	  p_value = pointer.getPtr()[*poin_num_p];
	  poin_num_level_p++;

      default:;
      }
    ++poin_num_p;
  }
  else if (r_action_p[i] < FI_VRBL) {
    if (r_action_p[i] != FI_SELF) {
      non_lvalue = false;
      p_value = &argument[r_action_p[i]];
      value = argument[r_action_p[i]];
    } else {
      int k = 0;
      double* l_number_p    = number_p;
      uint* l_var_num_p     = var_num_p;
      uint* l_poin_num_p    = poin_num_p;
      uint* l_poin_num_level_p     = poin_num_level_p;
      uint* l_usr_function_0_num_p = usr_function_0_num_p;
      uint* l_usr_function_1_num_p = usr_function_1_num_p;
      uint* l_usr_function_2_num_p = usr_function_2_num_p;
      uint* l_usr_function_2p_num_p = usr_function_2p_num_p;
      uint* l_usr_function_2d_num_p = usr_function_2d_num_p;
      offsets_t* l = offsets_p;
      number_p     = number.getPtr();
      var_num_p     = var_num.getPtr();
      poin_num_p    = poin_num.getPtr();
      poin_num_level_p    = poin_num_level.getPtr();
      usr_function_0_num_p = usr_function_0_num.getPtr();
      usr_function_1_num_p = usr_function_1_num.getPtr();
      usr_function_2_num_p = usr_function_2_num.getPtr();
      usr_function_2p_num_p = usr_function_2p_num.getPtr();
      usr_function_2d_num_p = usr_function_2d_num.getPtr();
      offsets_p = offsets.getPtr();

      ++rl();
      do {
	if (r_calculate(k, value, pval)) {
if (!locale) {
	  c.err(0,0,"forint internal error.\n");
} else {
	  c.err(0,0,"  forint.\n");
}
          c.flsh(); 
	  error_code = -101;
	  return error_code;
	}
      } while (++k < int(r_action.size()));
      --rl();

      number_p      = l_number_p;
      var_num_p     = l_var_num_p;
      poin_num_p    = l_poin_num_p;
      poin_num_level_p     = l_poin_num_level_p;
      usr_function_0_num_p = l_usr_function_0_num_p;
      usr_function_1_num_p = l_usr_function_1_num_p;
      usr_function_2_num_p = l_usr_function_2_num_p;
      usr_function_2p_num_p = l_usr_function_2p_num_p;
      usr_function_2d_num_p = l_usr_function_2d_num_p;
      offsets_p = l;
    }
  }
  else if (r_action_p[i] < FI_FUNC) {
    value = constant[r_action_p[i] - FI_EN_E];
  }
// an user defined function with an empty signature...
  else if (r_action_p[i] == FI_FUNC) {
    value = (*usr_function_0_p[*usr_function_0_num_p++])();
  }
// an user defined function with one argument...
  else if (r_action_p[i] == FI_FUNC1) {
    r_calculate(++i, val1, pval);
    value = (*usr_function_1_p[*usr_function_1_num_p++])(val1);
  }
// an user defined function with two doubles...
  else if (r_action_p[i] == FI_FUNC2D) {
    r_calculate(++i, val1, pval);
    r_calculate(++i, val2, pval2);
    value = (*usr_function_2d_p[*usr_function_2d_num_p++]) (val1, val2);
  }
// an user defined function with two pointers...
  else if (r_action_p[i] == FI_FUNC2P) {
    r_calculate(++i, val1, pval);
    r_calculate(++i, val2, pval2);
    value = (*usr_function_2p_p[*usr_function_2p_num_p++]) (pval, pval2);
  }
// an user defined function with pointer and double...
  else if (r_action_p[i] == FI_FUNC2) {
    r_calculate(++i, val1, pval);
    r_calculate(++i, val2, pval2);
    value = (*usr_function_2_p[*usr_function_2_num_p++]) (val1, pval2);
  }
  else if (r_action_p[i] < FI_OPER) {
    n = i;
    r_calculate(++i,val1, pval);
    if (r_action_p[n] < FI_ASIN) {
      if (angle_unit == FI_DEGREE) val1 *= M_PI/180.;
      else if (angle_unit == FI_GRAD) val1 *= M_PI/200.;
    }
    value = func_arr[r_action_p[n] - FI_SIN](val1);
    if (r_action_p[n] > FI_TAN && r_action_p[n] < FI_SINH) {
      if (angle_unit == FI_DEGREE) value *= 180./M_PI;
      else if (angle_unit == FI_GRAD) value *= 200./M_PI;
    }
  }
  else if (r_action_p[i] < FI_COND) {
    n = i;
    if (!cont_flag) r_calculate(++i, val1, pval);
    else val1 = value;
    r_calculate(++i, val2, pval);
    value = oper_arr[r_action_p[n] - FI_ADD](val1, val2);
  }
  else if (r_action_p[i] < FI_LOGC) {
    n = i;
    if (!cont_flag) r_calculate(++i, val1, pval);
    else val1 = value;
    r_calculate(++i, val2, pval);
    value = cond_arr[r_action_p[n] - FI_GT](val1, val2);
  }
  else if (r_action_p[i] < FI_GRAM) {
    n = i;
    if (!cont_flag) r_calculate(++i, val1, pval);
    else val1 = value;
    if (r_action_p[n] == FI_AND &&  val1 ||
	r_action_p[n] == FI_OR  && !val1) {
      r_calculate(++i, val2, pval);
      //value = logc_arr[r_action_p[n] - FI_AND](val1, val2);
      value = val2;
      ++offsets_p;
    } else {
      value = val1;
      i += offsets_p->ra;
      //skip_number = offsets_p->ra;
      make_offsets();
    }
  }
  else if (r_action_p[i] == FI_IF) {
    r_calculate(++i, val1, pval);
    cond = val1;
    if (cond) {
      ++offsets_p;
      r_calculate(++i, val1, pval);
      value = val1;
// jump...
      i += offsets_p->ra;
      make_offsets();
      //skip_number = offsets_p->ra;
    } else {
// jump...
      i += offsets_p->ra;
      make_offsets();
      r_calculate(++i, val1, pval);
      value = val1;
      ++offsets_p;
      //skip_number = offsets_p->ra;
    }
  }
  else if (r_action_p[i] >= FI_ASS) {
    n = i;
// get val1...
    if (!cont_flag) r_calculate(++i, val1, pval);
    else val1 = value;
// get pval...
    r_calculate(++i, val2, pval);
// and assign value to it...
    value = ass_arr[r_action_p[n]-FI_ASS]((double*)pval, val1);
  }
  else if (r_action_p[i] == FI_WHILE) {
    r_calculate(++i, val2, pval);
    while (val2 && !break_flag) {
      ++offsets_p;
      r_calculate(++i, val1, pval);
      i += offsets_p->ra + 1;
      make_offsets();
      r_calculate(++i, val2, pval);
    }
// jump...
    i += offsets_p->ra;
    make_offsets();
    value = val1;
  }
  if (i < int(r_action.size())-3 &&
      r_action_p[i+1] == FI_CONTINUE &&
      non_lvalue) {
    if (r_action_p[i+2] != FI_CONTINUE) {
      i += 2;
      cont_flag = true;
      goto cont;
    } else ++i;
  }
  else return 0;
  return 0;
}

// a stack virtual machine...
int forint::
s_calculate(int    i,       // start with i-th command in the "s_action"
	    double& value,  // returned value
	    int     length) // length of a command sequence to be executed
{
  using namespace FInt_internal;
  register double val1 = 0;
  register double val2 = 0;
  register void*  pval  = 0;
  register void*  pval2 = 0;
  register long j = 0;
  register char*   t1 = 0;
  register short*  t2 = 0;
  register int*    t3 = 0;
  register long*   t4 = 0;
  register long long* t41 = 0;
  register float*  t5 = 0;
  register double* t6 = 0;
  register long double* t7 = 0;

  register stack_t<double> stck;
  register stack_t<void*> pstck;

  for (; length-->0; i++) {
    if (s_action_p[i] == FI_NUMB) {
      stck.push(val1);
      val1 = *number_p++;
      continue;
    }
    if (s_action_p[i] == FI_VRBL) {
      stck.push(val1);
      pstck.push(pval);
      pval = variable_p[*var_num_p];
      val1 = *variable_p[*var_num_p++];
      continue;
    }
    if (s_action_p[i] == FI_POIN) {
      pstck.push(pval);
      uint* pp = poin_offset.getPtr()[*poin_num_p] +
	((*poin_num_level_p) ? *poin_num_level_p - 1 : 0);
      val1 = (*poin_num_level_p) ? val1 : 0;

      uint* ppp = poin_dim_size.getPtr()[*poin_num_p] +
	((*poin_num_level_p) ? *poin_num_level_p - 1 : 0);

      if (poin_type.getPtr()[*poin_num_p].level) {
	void** t = (void**)pointer.getPtr()[*poin_num_p];
	j = *poin_num_level_p++;
	while(ppp) {
	  if (val1 > *ppp || val1 < 0) {
	    out_of_bound(poin_name.getPtr()[*poin_num_p],*ppp,int(val1));
	    break;
	  }
	  ppp--;
	  t = t + (int)val1 * *pp--;
	  if (j && --j) val1 = stck.pop();
	  else break;
	}
	pval = t;
      } else
	switch (poin_type.getPtr()[*poin_num_p].b_type) {

	case FI_CHAR:
	  t1 = (char*)pointer.getPtr()[*poin_num_p];
	  j = *poin_num_level_p++;
	  while(ppp) {
	    if (val1 > *ppp || val1 < 0) {
	      out_of_bound(poin_name.getPtr()[*poin_num_p],*ppp,int(val1));
	      break;
	    }
	    ppp--;
	    t1 += (int)val1 * *pp--;
	    if (j && --j) val1 = stck.pop();
	    else break;
	  }
	  pval = t1;
	  val1 = *t1;
	  break;

	case FI_SHORT:
	  t2 = (short*)pointer.getPtr()[*poin_num_p];
	  j = *poin_num_level_p++;
	  while(ppp) {
	    if (val1 > *ppp || val1 < 0) {
	      out_of_bound(poin_name.getPtr()[*poin_num_p],*ppp,int(val1));
	      break;
	    }
	    ppp--;
	    t2 += (int)val1 * *pp--;
	    if (j && --j) val1 = stck.pop();
	    else break;
	  }
	  pval = t2;
	  val1 = *t2;
	  break;

	case FI_INT:
	  t3 = (int*)pointer.getPtr()[*poin_num_p];
	  j = *poin_num_level_p++;
	  while(ppp) {
	    if (val1 > *ppp || val1 < 0) {
	      out_of_bound(poin_name.getPtr()[*poin_num_p],*ppp,int(val1));
	      break;
	    }
	    ppp--;
	    t3 += (int)val1 * *pp--;
	    val1 = stck.pop();
	    if (j && --j) val1 = stck.pop();
	    else break;
	  }
	  pval = t3;
	  val1 = *t3;
	  break;

	case FI_LONG:
	  t4 = (long*)pointer.getPtr()[*poin_num_p];
	  j = *poin_num_level_p++;
	  while(ppp) {
	    if (val1 > *ppp || val1 < 0) {
	      out_of_bound(poin_name.getPtr()[*poin_num_p],*ppp,int(val1));
	      break;
	    }
	    ppp--;
	    t4 += (int)val1 * *pp--;
	    if (j && --j) val1 = stck.pop();
	    else break;
	  }
	  pval = t4;
	  val1 = *t4;
	  break;

	case FI_LONG_LONG:
	  t41 = (long long*)pointer.getPtr()[*poin_num_p];
	  j = *poin_num_level_p++;
	  while(ppp) {
	    if (val1 > *ppp || val1 < 0) {
	      out_of_bound(poin_name.getPtr()[*poin_num_p],*ppp,int(val1));
	      break;
	    }
	    ppp--;
	    t41 += (int)val1 * *pp--;
	    if (j && --j) val1 = stck.pop();
	    else break;
	  }
	  pval = t41;
	  val1 = *t41;
	  break;

	case FI_FLOAT:
	  t5 = (float*)pointer.getPtr()[*poin_num_p];
	  j = *poin_num_level_p++;
	  while(ppp) {
	    if (val1 > *ppp || val1 < 0) {
	      out_of_bound(poin_name.getPtr()[*poin_num_p],*ppp,int(val1));
	      break;
	    }
	    ppp--;
	    t5 += (int)val1 * *pp--;
	    if (j && --j) val1 = stck.pop();
	    else break;
	  }
	  pval = t5;
	  val1 = *t5;
	  break;

	case FI_DOUBLE:
	  t6 = (double*)pointer.getPtr()[*poin_num_p];
	  j = *poin_num_level_p++;
	  while(ppp) {
	    if (val1 > *ppp || val1 < 0) {
	      out_of_bound(poin_name.getPtr()[*poin_num_p],*ppp,int(val1));
	      break;
	    }
	    ppp--;
	    t6 += (int)val1 * *pp--;
	    if (j && --j) val1 = stck.pop();
	    else break;
	  }
	  pval = t6;
	  val1 = *t6;
	  break;

	case FI_LONG_DOUBLE:
	  t7 = (long double*)pointer.getPtr()[*poin_num_p];
	  j = *poin_num_level_p++;
	  while(ppp) {
	    if (val1 > *ppp || val1 < 0) {
	      out_of_bound(poin_name.getPtr()[*poin_num_p],*ppp,int(val1));
	      break;
	    }
	    ppp--;
	    t7 += (int)val1 * *pp--;
	    if (j && --j) val1 = stck.pop();
	    else break;
	  }
	  pval = t7;
	  val1 = *t7;
	  break;

	case FI_USR_CLASS:
	  pval = pointer.getPtr()[*poin_num_p];
	  poin_num_level_p++;

	default:;
	}
      ++poin_num_p;
      continue;
    }
    if (s_action_p[i] < FI_VRBL) {
      if (s_action_p[i] != FI_SELF) {
	stck.push(val1);
	pstck.push(pval);
	pval = &argument[s_action_p[i]];
	val1 = argument[s_action_p[i]];
      } else {
	uint k = 0;
        double* l_number_p    = number_p;
        uint* l_var_num_p     = var_num_p;
        uint* l_poin_num_p    = poin_num_p;
        uint* l_poin_num_level_p     = poin_num_level_p;
        uint* l_usr_function_0_num_p = usr_function_0_num_p;
        uint* l_usr_function_1_num_p = usr_function_1_num_p;
        uint* l_usr_function_2_num_p = usr_function_2_num_p;
        uint* l_usr_function_2p_num_p = usr_function_2p_num_p;
        uint* l_usr_function_2d_num_p = usr_function_2d_num_p;
	offsets_t* l = offsets_p;
        number_p     = number.getPtr();
        var_num_p     = var_num.getPtr();
        poin_num_p    = poin_num.getPtr();
        poin_num_level_p    = poin_num_level.getPtr();
        usr_function_0_num_p = usr_function_0_num.getPtr();
        usr_function_1_num_p = usr_function_1_num.getPtr();
        usr_function_2_num_p = usr_function_2_num.getPtr();
        usr_function_2p_num_p = usr_function_2p_num.getPtr();
        usr_function_2d_num_p = usr_function_2d_num.getPtr();
	offsets_p = offsets.getPtr();

	++rl();
	if (s_calculate(k, val1, s_action.size())) {
if (!locale) {
	  c.err(0,0,"forint internal error.\n");
} else {
	  c.err(0,0,"  forint.\n");
}
          c.flsh(); 
	  error_code = -100;
	  return error_code;
	}
	--rl();

        number_p     = l_number_p;
        var_num_p     = l_var_num_p;
        poin_num_p    = l_poin_num_p;
        poin_num_level_p     = l_poin_num_level_p;
        usr_function_0_num_p = l_usr_function_0_num_p;
        usr_function_1_num_p = l_usr_function_1_num_p;
        usr_function_2_num_p = l_usr_function_2_num_p;
        usr_function_2p_num_p = l_usr_function_2p_num_p;
        usr_function_2d_num_p = l_usr_function_2d_num_p;
	offsets_p = l;
      }
      continue;
    }
    if (s_action_p[i] < FI_FUNC) {
      stck.push(val1);
      val1 = constant[s_action_p[i] - FI_EN_E];
      continue;
    }
// an user defined function with an empty signature...
    if (s_action_p[i] == FI_FUNC) {
      stck.push(val1);
      val1 = (*usr_function_0_p[*usr_function_0_num_p++])();
      continue;
    }
// an user defined function with one argument...
    if (s_action_p[i] == FI_FUNC1) {
      val1 = (*usr_function_1_p[*usr_function_1_num_p++])(val1);
      continue;
    }
// an user defined function with two arguments...
    if (s_action_p[i] == FI_FUNC2D) {
      val2 = stck.pop();
      val1 = (*usr_function_2d_p[*usr_function_2d_num_p++]) (val2, val1);
      continue;
    }
// an user defined function with two arguments...
    if (s_action_p[i] == FI_FUNC2P) {
      pval2 = pstck.pop();
      val1 = (*usr_function_2p_p[*usr_function_2p_num_p++]) (pval2, pval);
      continue;
    }
// an user defined function with two arguments...
    if (s_action_p[i] == FI_FUNC2) {
      val2 = stck.pop();
      val1 = (*usr_function_2_p[*usr_function_2_num_p++]) (val2, pval);
      continue;
    }
    if (s_action_p[i] < FI_OPER) {
      if (s_action_p[i] < FI_ASIN) {
	if (angle_unit == FI_DEGREE) val1 *= M_PI/180.;
	else if (angle_unit == FI_GRAD) val1 *= M_PI/200.;
      }
      val1 = func_arr[s_action_p[i] - FI_SIN](val1);
      if (s_action_p[i] > FI_TAN && s_action_p[i] < FI_SINH) {
	if (angle_unit == FI_DEGREE) val1 *= 180./M_PI;
	else if (angle_unit == FI_GRAD) val1 *= 200./M_PI;
      }
      continue;
    }
    if (s_action_p[i] < FI_COND) {
      val2 = stck.pop();
      val1 = oper_arr[s_action_p[i] - FI_ADD](val2, val1);
      continue;
    }
    if (s_action_p[i] < FI_LOGC) {
      val2 = stck.pop();
      val1 = cond_arr[s_action_p[i] - FI_GT](val2, val1);
      continue;
    }
    if (s_action_p[i] < FI_GRAM) {
      //val2 = stck.pop();
      //val1 = logc_arr[s_action_p[i] - FI_AND](val2, val1);
      continue;
    }
    if (s_action_p[i] == FI_IF) {
      if (val1) {
	offsets_p++;
	continue;
      }
      else {
// jump...
	i += offsets_p->sa;
	length -= offsets_p->sa;
	make_offsets();
	offsets_p++;
	continue;
      }
    }
    if (s_action_p[i] == FI_ELSE) {
// jump...
      i += offsets_p->sa;
      length -= offsets_p->sa;
      make_offsets();
      continue;
    }
    if (s_action_p[i] == FI_WHILE) {
      if (!break_flag) {
// jump...
	i += offsets_p->sa;
	i--;
	length -= offsets_p->sa - 1;
	make_offsets();
	stck.cut();
	pstck.cut();
      }
      continue;
    }
    if (s_action_p[i] == FI_JUMP) {
      if ((s_action_p[i + offsets_p->sa] == FI_AND && !val1 ||
	   s_action_p[i + offsets_p->sa] == FI_OR  &&  val1)) {
// omit && or || operator...
	i += offsets_p->sa;
	length -= offsets_p->sa;
	make_offsets();
      } else ++offsets_p;
      continue;
    }
    if (s_action_p[i] >= FI_ASS) {
      val1 = ass_arr[s_action_p[i]-FI_ASS]((double*)pval, stck.pop());
      continue;
    }
  }
  value = val1;
  return 0;
}

// calculate by a dynamically linked function...
int forint::
dl_calculate(double& value,  // returned value
	     char* expr)
{
  int rv = 0;

  if (!dlfunc_lib) {
      fputs ("\nforint: DLF library hasn't been opened (Probably call"
	     " of \'makeDLF()\' hasn't been made or failed).\n", stderr);
      rv = error_code = 1010;
      return error_code;
  }
  value = (*dl_function)();
  return rv;
}

void forint::
out_of_bound(char* name, uint top, int index)
{
if (!locale) {
    c.err(0,0,"Overrun of array \'");
    c.err(name,5);
    c.err("\' bounds [");
    c.err("0",5);
    c.err(",");
    c.err(double(top),5);
    c.err("], try to access by index ");
    c.err(double(index),5);
    c.err(".\n");
    c.flsh(); 
} else {
    c.err(0,0,"   [");
    c.err("0",5);
    c.err(",");
    c.err(double(top),5);
    c.err("]  \'");
    c.err(name,5);
    c.err("\',     ");
    c.err(double(index),5);
    c.err(".\n");
    c.flsh(); 
}
}

}
